#!/usr/bin/env ruby
#
# Google Chrome Stable updates blog scraper
#
# Download "Google Chrome Releases" blog pages into individual posts
#
# Author: Joshua J. Drake / Accuvant LABS
# License: BSD
#

require 'nokogiri'

require 'date'

require 'net/http'
require 'uri'


#
# Scrape some data from all posts in the Google Chrome blog local copy
#
def get_posts()
	ret = []
	page = 0

	loop {
		html = nil
		fname = "pages/Stable_updates_pg#{page}.html"
		$stderr.puts "*** Processing #{fname} ..." if $debug
		break if not File.exists?(fname)
		File.open(fname, "rb") { |fd|
			html = fd.read
		}

		doc = Nokogiri::HTML(html)
		#puts doc.inspect

		# Process posts
		new_post = nil
		doc.xpath('//div[@class = "post"]').each { |node|
			new_post = {}
			node.xpath('h3/a').each { |c|
				raise 'Already found link!' if new_post[:href]
				new_post[:href] = c['href']
			}

			node.xpath('p[@class = "post-subhead"]').each { |c|
				next if c.text =~ /Labels:/
				raise 'Already found date!' if new_post[:date]
				date_text = c.text.strip.gsub(/ \| /, ' ')
				new_post[:date] = Date.parse(date_text)
			}

			if not new_post[:vers]
				txt = node.text.strip
				if txt =~ /([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)[,.\n ]/
					new_post[:vers] = $1
				elsif new_post[:date] == Date.parse('2009-10-05')
					new_post[:vers] = '3.0.195.25'
				end
			end

			[ :vers, :href, :date ].each { |k|
				raise "missing key #{k.to_s}" if not new_post.has_key? k
			}

			# These apparently have no security consequence, so we skip them
			ver = new_post[:vers]
			if (not ver) or ([
					'11.0.696.60',   # only a functionality regression
					'10.0.648.151',  # comodo cert fiasco blockage
					'9.0.597.98',    # only a functionality regression
					'8.0.552.231',   # "stability fix" for web forms on mac
					'6.0.472.55',    # only fixes for functionality
					'4.1.249.1042',  # fixes extension load problem only
					'3.0.195.38',    # only crashes, no security fixes
					'3.0.195.25',    # adds image link to new tab page (themes)
					'2.0.172.28',    # initial release of 2.0, no security fixes mentioned
					'1.0.154.53',    # no security fixes
					'1.0.154.43',    # gears crash, but no bug reference
					'1.0.154.36',    # initial stable release

					# The following contain only flash fixes
					'12.0.742.100',
					'11.0.696.65',
					'11.0.696.77',
					'10.0.648.134',
					'5.0.375.126',
					].include? ver)
				#$stderr.puts "Eek, no security changes found?"
				next
			end

			#puts new_post.inspect
			ret << new_post
		}

		page += 1
	}
	ret
end



posts = get_posts()

posts.sort! { |a,b| a[:date] <=> b[:date] }

# Turn the version numbers into milestone numbers
milestones = {}
posts.each { |p|
	ver = p[:vers]
	next if not ver

	parts = ver.split('.').map { |el| el.to_i }
	ms = 'm%02d.%d' % [ parts[0], parts[1] ]

	# Increment the global counter for each milestone
	milestones[ms] ||= 0
	p[:ms] = "#{ms}su#{milestones[ms]}"
	milestones[ms] += 1
}

posts.sort! { |a,b| a[:ms] <=> b[:ms] }

# Download the individual pages...
opts = {}
posts.each { |post|
	fname = post[:ms] + '.html'

	next if File.exists?(fname)

	puts "%s  %-13s %s" % [ post[:ms], post[:vers], post[:href] ]

	# Download it...
	uri = URI.parse(post[:href])
	req = Net::HTTP::Get.new(uri.request_uri, opts)
	res = Net::HTTP.start(uri.host, uri.port) { |http|
		http.request(req)
	}

	# Save it..
	$stderr.puts "*** Saving contents to #{fname}"
	File.open(fname, "wb") { |fd|
		fd.write res.body
	}
}
