#!/usr/bin/env ruby
#
# Mozilla Firefox Advisory scraper
#
# Author: Joshua J. Drake / Accuvant LABS
# License: BSD
#
#
# Mozilla data is available under a Creative Commons License. For
# more information, see:
#
# http://www.mozilla.org/foundation/licensing/website-content.html
#
#
# NOTES:
#
# The following advisories were manually fixed to include "<h3>References</h3>" 
# before the credits/refernces section starts.
#
# MFSA 2009-64
# MFSA 2011-17
# MFSA 2011-18
#
# The following advisories have the CVE number placed next to the wrong bug
# number.
#
# MFSA 2008-35
# MFSA 2008-37
# MFSA 2009-12
#
# Unclear which bug the CVE belongs to:
#
# MFSA 2008-39
#
# Although MFSA 2011-17 includes one ref to 623791 and one to 623791#c6, 
# these are both distinct vulns (with distinct CVEs).
#
# MFSA 2010-17 doesn't have a title for the bug list so the subtitle
# doesn't get used because the buglist modified :href differs from the
# original :href which matches :text
#


require 'date'

require 'nokogiri'

require 'mysql'
$db = Mysql::new("localhost", "root", "", "browsers")

#$debug = true

def insert_vuln(adv_id, title, vendor_id, cve, impact, reported, discovered, disclosure)
	values = "'1'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(title) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(vendor_id) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(cve) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(impact) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(reported) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(disclosure) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(discovered) + "'"
	values << ", " if values.length > 0
	values << "'" + adv_id.to_s + "'"
	
	qstr = "insert into vulns (prod_id, title, vendor_id, cve, severity, report_date, disclosure_date, discover_date, advisory_id)"
	qstr << " values (" << values << ")"
	$db.query(qstr)
end

def insert_adv(title, vendor_id, publish)
	values = "'1'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(title) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(vendor_id) + "'"
	values << ", " if values.length > 0
	values << "'" + $db.escape_string(publish) + "'"
	
	qstr = "insert into advisories (prod_id, title, vendor_id, publish_date)"
	qstr << " values (" << values << ")"
	$db.query(qstr)
	return $db.insert_id()
end


def show_adv(adv)
	prod = adv[:prod].join(', ')
	if prod !~ /Firefox/i
		$stderr.puts "Firefox is not affected: " + prod
		return
	end

	puts "=" * 80
	puts "#{adv[:title]} @ #{adv[:date]}"
	# adv[:longname]
	puts "Subtitle: " + adv[:subtitle].inspect
	puts "Impact: " + adv[:impact].inspect
	puts "Products: " + prod
	puts "Fixed in: " + adv[:fix].inspect
	puts "Finder: " + adv[:finder].inspect
	if adv[:body]
		puts ""
		puts "Description: " + adv[:body].join("\n\n")
		puts ""
	end
	if adv[:workaround]
		puts "Workaround: " + adv[:workaround].join("\n\n")
		puts ""
	end

	if (not (adv[:title].include? adv[:subtitle])) and $debug
		$stderr.puts "Title #{adv[:title].inspect} doesn't include subtitle #{adv[:subtitle].inspect}!"
	end

	title_parts = adv[:title].split(':')
	mfsa_id = title_parts.first
	title = title_parts[1, title_parts.length].join(':')

	adv_id = insert_adv(title, mfsa_id, adv[:date].to_s)

	# Last minute fix ups for a special cases.
	case mfsa_id
	when 'MFSA 2010-33'
		adv[:vulns].each { |v|
			if v[:bug] == '577512'
				v[:date] = Date.parse("October 27, 2010")
			end
		}
	end


	adv[:vulns].each_with_index { |vuln,vulnidx|
		cve = ""

		puts "Vuln # #{vulnidx+1}:"
		if vuln[:cve] and vuln[:cve].length > 0
			cve = vuln[:cve].join(', ')
			puts "  CVE(s): " + cve
		else
			puts "  CVE: ?"
		end
		puts "  BUG: " + (vuln[:bug] || "?")
		if vuln[:text] != vuln[:href]
			puts "  TXT: " + (vuln[:text] || "?")
		end
		puts "  URL: " + (vuln[:href] || "?")
		if vuln[:body] and vuln[:body].length > 0
			puts "  BODY: " + vuln[:body].join("\n\n")
		else
			puts "  BODY: ?"
		end

		# insert vuln
		bug_title = "%s # %d" % [adv[:subtitle], vulnidx+1]
		if vuln[:text] and vuln[:text] != vuln[:href]
			bug_title = vuln[:text]
		end
		date = vuln[:date] || adv[:date]
		insert_vuln(adv_id,
			bug_title, vuln[:bug].to_s,
			cve,
			adv[:impact].to_s,
			'', '', date.to_s)
	}
	puts ""
end


def append_vuln(adv, vuln, buglist)
	if buglist.length > 0
		buglist.each { |bug|
			v = vuln.clone
			v[:bug] = bug
			v[:href] = 'https://bugzilla.mozilla.org/show_bug.cgi?id=' + bug.to_s
			$stderr.puts "  appending: #{v.inspect}" if $debug
			adv[:vulns] << v
		}
		buglist.clear
	else
		# If we ever try to add something with only a body, it must be an error.
		# Instead, add this onto the advisory body.
		if vuln.keys.length == 1 and vuln.keys.include? :body
			adv[:body] ||= []
			adv[:body] << vuln[:body]
		else
			$stderr.puts "  appending: #{vuln.inspect}" if $debug
			adv[:vulns] << vuln
		end
	end
	{}
end


#
# Scrape a single adv
#
def scrape_adv(node, adv)
	adv[:vulns] = []
	vuln = {}
	buglist = []

	section = nil
	node.children.each { |c|
		txt = c.text.strip
		# Skip empty text stuff
		next if c.name == 'text' and txt.empty?

		case c.name
		when 'p'
			$stderr.puts "found <p>" if $debug
			
			# Skip the breadcrumb paragraph
			next if txt =~ /^You are here/

			# Grab description, workaround, etc text
			if txt !~ /^Title:/
				case section
				when 'Description', nil
					adv[:body] ||= []
					adv[:body] << txt.gsub("\n", "")

				when 'Workaround'
					adv[:workaround] ||= []
					adv[:workaround] << txt.gsub("\n", "")

				when 'References', 'Credits'
					# Body should always precede any references. Otherwise,
					# we probably have the start of a new vuln.
					if vuln.keys.length > 0
						$stderr.puts "appending vuln(s): encountered body" if $debug
						vuln = append_vuln(adv, vuln, buglist)
					end

					vuln[:body] ||= []
					vuln[:body] << txt.gsub("\n", "")

				else
					$stderr.puts "Found <p> in invalid section: #{section.inspect}"
					#$stderr.puts txt.inspect

				end

				next
			end

			#puts "found <p>: #{c.inspect}"
			#puts "text: #{txt}"

			key = nil
			allow_mult = nil
			c.children.each { |lns|
				txt = lns.text.strip
				# we skip <2 since there are some unicode chars (from &nbsp;)
				next if txt.length < 2

				# Fix up some span stuff
				if txt =~ /span class/
					txt = txt.split(/<|>/)[4].strip
				end
				$stderr.puts txt.inspect if $debug

				if txt =~ /:$/
					case txt
					when "Title:"
						key = :subtitle
						allow_mult = false

					when "Impact:"
						key = :impact
						allow_mult = false
						
					when "Announced:"
						key = :date
						allow_mult = false
						
					when "Products:"
						key = :prod
						allow_mult = true
						
					when "Fixed in:"
						key = :fix
						allow_mult = true

					when "Reporter:"
						key = :finder
						allow_mult = true

					else
						# Uncomment to spot check strange labels
						$stderr.puts "Unknown label: #{txt}"

					end
				else
					#$stderr.puts "key: #{key.to_s}, allow_mult: #{allow_mult.inspect}" if $debug
					raise 'no key?' if not key

					case key
					when :date
						adv[key] = Date.parse(txt)
						#$stderr.puts "assigning #{txt.inspect} to key \"#{key.inspect}\"" if $debug

					else
						if allow_mult
							#$stderr.puts "appending #{txt.inspect} to key \"#{key.inspect}\"" if $debug
							adv[key] ||= []
							adv[key] << txt
						else
							if adv[key]
								raise "Eek, duplicate entry for key \"#{key.to_s}\""
							end
							#$stderr.puts "assigning #{txt.inspect} to key \"#{key.inspect}\"" if $debug
							adv[key] = txt
						end

					end
				end
			}

		when 'ul'
			$stderr.puts "found <ul>" if $debug

			# Scrape the references to figure out what vulns (more granular than cve) exist
			c.children.each { |ref|
				txt = ref.text.strip
				next if txt.empty?

				$stderr.puts "processing #{ref.name} - #{txt.length} bytes" if $debug

				case ref.name
				when 'li'
					raise "<li> encountered in section: #{section}" if not [ "References", "Credits" ].include? section
					anchor = nil
					ref.children.each { |ch|
						if ch.name == "a"
							raise 'multiple links in the reference!' if anchor
							anchor = ch
						end
					}
					raise 'reference missing anchor!' if not anchor
					atxt = anchor.text.strip

					$stderr.puts "anchor: #{anchor.inspect}" if $debug
					
					if anchor['href'] =~ /^https?:\/\/bugzilla\.mozilla\.org\/([^\?]*)\?.*=(.*)/
						# Depending on the CGI name, we process this one differently
						cgi = $1
						bugs = $2
						#$stderr.puts "cgi: #{cgi.inspect}, bugs: #{bugs.inspect}" if $debug

						# If we already have an href, we probably have multiple successive vulns listed
						# in the same unordered list
						if vuln[:href]
							$stderr.puts "appending vuln(s): second bugzilla href found" if $debug
							vuln = append_vuln(adv, vuln, buglist)
						end

						vuln[:href] = anchor['href']
						raise "Eek, vuln already has text!" if vuln[:text]
						vuln[:text] = atxt

						# Now process the bug nubmers from the cgi
						case cgi
						when 'show_bug.cgi'
							# If we already have a bug, we need to convert to a buglist
							if vuln[:bug]
								# If the title is different, this is probably a different bug
								if vuln[:text] != atxt
									$stderr.puts "appending vuln(s): bug with different link text" if $debug
									vuln = append_vuln(adv, vuln, buglist)
									vuln[:bug] = bugs
								else
									buglist << vuln[:bug]
									vuln[:bug] = nil # not used anyway
									buglist << bugs
								end
							else
								vuln[:bug] = bugs
							end

						when 'buglist.cgi'
							# Got a list of bugs when one was alerady defined? almost suredly
							# this means a new set of bugs...
							if buglist.length > 0
								$stderr.puts "appending vuln(s): buglist found when not empty" if $debug
								vuln = append_vuln(adv, vuln, buglist)
							end
							buglist = bugs.split(',')

						else
							$stderr.puts "Unknown bugzilla cgi encountered: #{cgi}"

						end

					elsif anchor['href'] =~ /^http:\/\/cve\.mitre\.org\/cgi-bin\/cvename\.cgi/ or
						anchor['href'] =~ /^http:\/\/web\.nvd\.nist\.gov\/view\/vuln\/detail\?vulnId=/
						vuln[:cve] ||= []
						vuln[:cve] << anchor.text

					else
						# Uncomment to spot check strange references
						#tmp = anchor.text.strip
						#tmp << " (" + anchor['href'] + ")"  if anchor['href']
						#$stderr.puts "Unknown reference: #{tmp}"

					end

				when 'comment'
					# ignore these, they appear to indicate missing cve refs
					#$stderr.puts "found <comment>: #{ref.inspect}"

				else
					$stderr.puts "Unknown element in <ul>: #{ref.name}"

				end
			}

		when 'ol'
			$stderr.puts "found <ol> in section #{section}" if $debug
			c.children.each { |el|
				txt = el.text.strip
				next if txt.empty?

				$stderr.puts "processing #{el.name} - #{txt.length} bytes" if $debug

				case el.name
				when 'li'
					case section
					when "Description"
						adv[:body] ||= []
						adv[:body] << txt.gsub("\n", "")

					when "Workaround"
						adv[:workaround] ||= []
						adv[:workaround] << txt.gsub("\n", "")

					else
						$stderr.puts "Found <ol> in unknown section: #{section}"

					end

				when 'text'
					# ignore, these are already processed with 'li' elements

				else
					$stderr.puts "Unknown element in <ol>: #{el.name}"

				end
			}

		when 'h1'
			# Long name
			if adv[:longname]
				raise "Eek, multiple long names found!"
			else
				$stderr.puts "Found long name: #{c.inspect}" if $debug
				adv[:longname] = c.text
			end

		when 'h3'
			$stderr.puts "Found h3: #{c.text}" if $debug
			section = c.text


		else
			# Uncomment to spot check unknown elements
			$stderr.puts "Unknown element: #{c.inspect}"

		end
	}

	if vuln.keys.length > 0
		$stderr.puts "appending vuln(s): stuff left at end" if $debug
		append_vuln(adv, vuln, buglist)
	end

	$stderr.puts "\n" + adv.inspect + "\n\n\n" if $debug
end


#
# Scrape all advisories from the MFSA data store
#
def get_advisories()
	ret = []
	
	fns = []
	Dir.glob('**/*.html') { |f|
		fns << f
	}

	fns.each { |f|
		html = nil
		$stderr.puts "*** Processing: #{f}" if $debug
		File.open(f, "rb") { |fd|
			html = fd.read
		}

		doc = Nokogiri::HTML(html)
		#puts doc.inspect

		# Process advisories
		new_adv = {}
		doc.xpath('//title').each { |node|
			raise 'title already found' if new_adv[:title]
			new_adv[:title] = node.text
		}
		doc.xpath('//div[@id="main-content"]').each { |node|
			scrape_adv(node, new_adv)
			#show_adv(new_adv)
			ret << new_adv
		}
		#break
	}
	ret
end


advisories = get_advisories()

advisories.sort! { |a,b| a[:date] <=> b[:date] }

advisories.each { |adv|
	show_adv(adv)
}
